import numpy as np
import matplotlib.pyplot as plt
from collections import namedtuple
from scipy.optimize import root


class LabIncompressible:
    def __init__(self, material):

        self.material = material
        self.title = self._get_title()

    def _get_title(self):
        return "Material Formulation: " + "-".join(
            [m.title() for m in self.material.fun.__name__.split("_")]
        )

    def _ux(self, stretch):
        "Principal stretches of incompressible uniaxial load case."
        return stretch, 1 / np.sqrt(stretch), 1 / np.sqrt(stretch)

    def _bx(self, stretch):
        "Principal stretches of incompressible equi-biaxial load case."
        return stretch, stretch, 1 / stretch**2

    def _ps(self, stretch):
        "Principal stretches of incompressible planar shear load case."
        return stretch, 1, 1 / stretch

    def _loadcase(self, stretch, kinematics):
        "Generalized load case `UX/BX/PS (Incompressible)`."

        def stress(stretch):
            "Evaluate the first Piola-Kirchhoff stress tensor."
            F = np.diag([*kinematics(stretch)])
            return self.material.gradient([F])[0]

        def stress_free(stretch):
            """First Piola-Kirchhoff stress tensor with resolved hydrostatic
            pressure due to incompressibility."""
            P = stress(stretch)
            return P[0, 0] - kinematics(stretch)[-1] / stretch * P[-1, -1]

        def stability(stretch):
            F = np.diag([*kinematics(stretch)])
            G = np.diag([*kinematics(stretch + 1e-6)])

            P = self.material.gradient([F])[0]
            A = self.material.hessian([F])[0]

            # convert hessian to (3, 3) matrix and take (2, 2) submatrix
            B = np.zeros((2, 2))
            delta = np.eye(2)

            for i in range(2):
                for j in range(2):
                    B[i, j] = (
                        A[i, i, j, j]
                        - delta[-1, j] / kinematics(stretch)[i] * P[-1, -1]
                        - kinematics(stretch)[-1]
                        / kinematics(stretch)[i]
                        * A[-1, -1, j, j]
                        + kinematics(stretch)[-1]
                        / kinematics(stretch)[i] ** 2
                        * P[-1, -1]
                        * delta[i, j]
                    )

            # init unit force in direction 1
            df = np.zeros(2)
            df[0] = 1

            # calculate linear solution of stretch 1 resulting from unit load
            dl = (np.linalg.inv(B) @ df)[0]

            # check slope of force
            Q = self.material.gradient([G])[0][0, 0]
            P = self.material.gradient([F])[0][0, 0]

            return True if dl > 0 and (Q - P) > 0 else False

        return (
            stress_free(stretch),
            *kinematics(stretch)[1:],
            stability(stretch),
        )

    def _shear(self, shear):
        "Load case `Simple Shear (Incompressible)`."

        def stress(shear):
            F = np.eye(3)
            F[0, 1] = shear
            return self.material.gradient([F])[0]

        return (
            stress(shear)[0, 1],
            1,
            1,
            None,
        )

    def run(
        self,
        ux=True,
        bx=True,
        ps=True,
        shear=True,
        stretch_min=None,
        stretch_max=2.5,
        shear_max=1.0,
        num=50,
    ):
        "Run load cases `UX/BX/PS/Simple-Shear (Incompressible)`."

        out = []

        Data = namedtuple(
            "Data", "label stretch stretch_2 stretch_3 shear stress stability"
        )

        if ux:
            if stretch_min is None:
                stretch_min = max(0, 1 - (stretch_max - 1) / 5)

            stretch_ux = np.linspace(stretch_min, stretch_max, num)
            stress_ux, stretch_2_ux, stretch_3_ux, stability_ux = np.array(
                [self._loadcase(s11, kinematics=self._ux) for s11 in stretch_ux]
            ).T
            ux_data = Data(
                "uniaxial",
                stretch_ux,
                stretch_2_ux,
                stretch_3_ux,
                0,
                stress_ux,
                stability_ux,
            )
            out.append(ux_data)

        if bx:
            stretch_bx = np.linspace(1, (stretch_max - 1) / 2 + 1, num)
            stress_bx, stretch_2_bx, stretch_3_bx, stability_bx = np.array(
                [self._loadcase(s11, kinematics=self._bx) for s11 in stretch_bx]
            ).T
            bx_data = Data(
                "biaxial",
                stretch_bx,
                stretch_2_bx,
                stretch_3_bx,
                0,
                stress_bx,
                stability_bx,
            )
            out.append(bx_data)

        if ps:
            stretch_ps = np.linspace(1, stretch_max, num)
            stress_ps, stretch_2_ps, stretch_3_ps, stability_ps = np.array(
                [self._loadcase(s11, kinematics=self._ps) for s11 in stretch_ps]
            ).T
            ps_data = Data(
                "planar",
                stretch_ps,
                stretch_2_ps,
                stretch_3_ps,
                0,
                stress_ps,
                stability_ps,
            )
            out.append(ps_data)

        if shear:
            deformation_sh = np.linspace(0, shear_max, num)
            stress_shear, stretch_2_sh, stretch_3_sh, stability_sh = np.array(
                [self._shear(g) for g in deformation_sh]
            ).T
            shear_data = Data(
                "shear",
                1,
                1,
                1,
                deformation_sh,
                stress_shear,
                stability_sh,
            )
            out.append(shear_data)

        return out

    def plot(self, data, stability=False):
        "Plot results of UX/BX/PS load cases."

        fig, ax = plt.subplots()

        lineargs = {
            "uniaxial": {"color": "C0"},
            "biaxial": {"color": "C1"},
            "planar": {"color": "C2"},
        }

        if data[-1].label == "shear":
            data = data[:-1]

        for d in data:

            if stability:

                stable = np.array(d.stability, dtype=bool)

                unstable_idx = np.arange(len(d.stretch))[~stable]
                stable_idx = np.arange(len(d.stretch))[stable]

                stress_stable = d.stress.copy()
                stress_stable[unstable_idx] = np.nan

                stress_unstable = d.stress.copy()
                stress_unstable[stable_idx] = np.nan

                ax.plot(d.stretch, stress_stable, **lineargs[d.label], label=d.label)
                ax.plot(
                    d.stretch,
                    stress_unstable,
                    **lineargs[d.label],
                    linestyle="--",
                )

            else:

                ax.plot(d.stretch, d.stress, **lineargs[d.label], label=d.label)

        ax.grid()
        ax.set_title(self.title, fontsize=10)
        ax.set_xlabel(r"stretch $\lambda_1 \quad \longrightarrow$")
        ax.set_ylabel(r"force per undeformed area $P_{11} \quad \longrightarrow$")
        ax.legend()

        fig.tight_layout()

        return fig, ax

    def plot_shear(self, data):
        "Plot results of shear load case."

        fig, ax = plt.subplots()

        lineargs = {
            "shear": {"color": "C3"},
        }

        if data[-1].label == "shear":
            d = data[-1]
        else:
            raise TypeError("No shear data found.")

        ax.plot(d.shear, d.stress, **lineargs[d.label], label=d.label)

        ax.grid()
        ax.set_title(self.title, fontsize=10)
        ax.set_xlabel(r"shear deformation $F_{12}=\gamma \quad \longrightarrow$")
        ax.set_ylabel(r"force per undeformed area $P_{12} \quad \longrightarrow$")
        ax.legend()

        fig.tight_layout()

        return fig, ax



from casadi import (
    # tensor and vector operations
    det,
    inv,
    dot as _dot,
    transpose,
    trace,
    diag,
    triu,
    tril,
    adj,
    cofactor,
    cross,
    times,
    eig_symbolic,
    qr,
    ldl,
    # trig
    sin,
    cos,
    tan,
    sinh,
    cosh,
    tanh,
    asin,
    acos,
    atan,
    atan2,
    asinh,
    acosh,
    atanh,
    # math
    exp,
    log,
    sqrt,
    sum1,
    sum2,
    sumsqr,
    pi,
    fabs,
    linspace,
    erf,
    erfinv,
    norm_1,
    sign,
    fmin,
    fmax,
    mmin,
    mmax,
    find,
    #
    if_else,
    logic_and,
    logic_or,
    logic_not,
    floor,
    ceil,
    SX,
    DM,
    MX,
    #
    vertcat,
    horzcat,
    vertsplit,
    horzsplit,
    repmat,
    reshape,
    gradient,
    hessian,
    Function,
)

eye = SX.eye
ones = SX.ones
zeros = SX.zeros


def zeros_like(T):

    return zeros(T.shape)


def ones_like(T):

    return ones(T.shape)


def invariants(T):

    I1 = trace(T)
    I2 = (I1**2 - trace(T @ T)) / 2
    I3 = det(T)

    return I1, I2, I3


def eigvals(T, eps=1e-4):

    D = DM([[1, 0, 0], [0, -1, 0], [0, 0, 0]])

    return eig_symbolic(T + D * eps)


def cof(T):

    return det(T) * transpose(inv(T))


def sym(T):

    return (T + transpose(T)) / 2


def dot(A, B):

    return _dot(transpose(A), B)


def dev(T):

    dim = T.shape[0]

    return T - trace(T) / dim * eye(dim)


def ddot(A, B):

    return trace(transpose(A) @ B)


def tresca(C):
    "Tresca Invariant as maximum difference of two eigenvalues."
    wC = eigvals(C, 8e-5)
    return mmax(fabs(wC[[0, 1, 2]] - wC[[1, 2, 0]]))


def mexp(C, eps=8e-5):
    "Exponential Function of a Matrix."
    w = eigvals(C, eps=eps)
    I = SX.eye(3)

    M1 = (C - w[1] * I) * (C - w[2] * I) / (w[0] - w[1]) / (w[0] - w[2])
    M2 = (C - w[2] * I) * (C - w[0] * I) / (w[1] - w[2]) / (w[1] - w[0])
    M3 = (C - w[0] * I) * (C - w[1] * I) / (w[2] - w[0]) / (w[2] - w[1])

    return exp(w[0]) * M1 + exp(w[1]) * M2 + exp(w[2]) * M3


def asvoigt(A, scale=1):

    if A.shape == (3, 3):
        return vertcat(
            A[0, 0],
            A[1, 1],
            A[2, 2],
            A[0, 1] * scale,
            A[1, 2] * scale,
            A[0, 2] * scale,
        )

    elif A.shape == (2, 2):
        return vertcat(
            A[0, 0],
            A[1, 1],
            A[0, 1] * scale,
        )

    else:
        raise ValueError("Unknown shape of input.")


def astensor(A, scale=1):

    if A.shape == (6, 1):
        A0 = vertcat(A[0] / scale, A[3] / scale, A[5])
        A1 = vertcat(A[3] / scale, A[1], A[4] / scale)
        A2 = vertcat(A[5] / scale, A[4] / scale, A[2])
        return horzcat(A0, A1, A2)

    elif A.shape == (3, 1):
        A0 = vertcat(A[0], A[2] / scale)
        A1 = vertcat(A[2] / scale, A[1])
        return horzcat(A0, A1)

    else:
        raise ValueError("Unknown shape of input.")



import numpy as np

from matadi import Variable, MaterialTensor
from matadi.models import (
    displacement_pressure_split,
    neo_hooke,
    ogden_roxburgh,
    morph,
    volumetric,
)

from matadi.math import det, gradient, dev, inv, asvoigt, astensor, vertcat, triu, zeros

import pytest


@displacement_pressure_split
def fun_nh_or(x, C10=0.5, bulk=5000, r=3, m=1, beta=0):
    "Strain energy function: Neo-Hooke & Ogden-Roxburgh."

    # split `x` into the deformation gradient and the state variable
    F, Wmaxn = x[0], x[-1]

    # isochoric and volumetric parts of the hyperelastic strain energy function
    W = neo_hooke(F, C10)
    U = volumetric(det(F), bulk)

    # pseudo-elastic softening
    eta, Wmax = ogden_roxburgh(W, Wmaxn, r, m, beta)

    # softened first Piola-Kirchhoff stress and updated state variable
    return eta * gradient(W, F) + gradient(U, F), Wmax


@displacement_pressure_split
def fun_morph(x, p1=0.035, p2=0.37, p3=0.17, p4=2.4, p5=0.01, p6=6.4, p7=5.5, p8=0.24):
    "Strain energy function: Neo-Hooke & Ogden-Roxburgh."

    # split `x` into the deformation gradient and the state variable
    F = x[0]

    P, statevars = morph(x, p1, p2, p3, p4, p5, p6, p7, p8)
    U = volumetric(det(F), 5000 * 2 * (p1 + p2))

    # first Piola-Kirchhoff stress and updated state variable
    return P + gradient(U, F), statevars


def test_up_state():

    # deformation gradient
    F = Variable("F", 3, 3)

    # state variables
    statevars = [Variable("z", 1, 1), Variable("z", 13, 1)]
    functions = [fun_nh_or, fun_morph]

    for fun, z in zip(functions, statevars):

        # get pressure variable from augmented function
        p = fun.p

        # Material as a function of `F` and `p`
        # with additional state variables `z`
        M = MaterialTensor([F, p, z], fun, triu=True, statevars=1)

        FF = (np.random.rand(3, 3, 8, 100) - 0.5) / 2
        pp = np.random.rand(1, 8, 100)
        zz = np.random.rand(*z.shape, 8, 100)

        for a in range(3):
            FF[a, a] += 1

        P = M.gradient([FF, pp, zz])  # stress, constraint, statevars_new
        A = M.hessian([FF, pp, zz])

        assert len(P) == 3
        assert len(A) == 3


def test_up_basic():

    # deformation gradient
    F = Variable("F", 3, 3)

    @displacement_pressure_split
    def fun(x):
        F = x[0]
        C = F.T @ F

        # (begin) test `asvoigt()` and `astensor()`
        C_3D = F.T @ F
        C_2D = vertcat(C_3D[0, 0], C_3D[1, 0], C_3D[0, 1], C_3D[1, 1]).reshape((2, 2))

        C_6 = asvoigt(C_3D)
        C_4 = asvoigt(C_2D)

        C_from_C_6 = astensor(C_6)
        C_2D_from_C_4 = astensor(C_4)

        assert C_3D[0, 1] == C_from_C_6[0, 1]
        assert C_2D[0, 1] == C_2D_from_C_4[0, 1]

        with pytest.raises(ValueError):
            asvoigt(C_6)

        with pytest.raises(ValueError):
            astensor(C_3D)
        # (end) test `asvoigt()` and `astensor()`

        return det(F) ** (-2 / 3) * dev(C) @ inv(C)

    # get pressure variable from augmented function
    p = fun.p

    # Material as a function of `F` and `p`
    M = MaterialTensor([F, p], fun, triu=True)

    FF = np.random.rand(3, 3, 8, 100)
    pp = np.random.rand(1, 8, 100)

    for a in range(3):
        FF[a, a] += 1

    P = M.gradient([FF, pp])  # stress, constraint
    A = M.hessian([FF, pp])

    assert len(P) == 2
    assert len(A) == 3


if __name__ == "__main__":
    test_up_basic()
    test_up_state()



from functools import wraps
from copy import deepcopy

from .. import Variable
from ..math import det, cof, trace, gradient


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient. Optionally, if
    the bulk keyword is passed, add a volumetric term."""

    @wraps(fun)
    def apply_iso(*args, **kwargs):
        F = args[0]
        J = det(F)
        F_iso = J ** (-1 / 3) * F

        fun_args = args[1:]
        fun_kwargs = deepcopy(kwargs)

        if "bulk" in kwargs.keys():
            _ = fun_kwargs.pop("bulk")

        W = fun(F_iso, *fun_args, **fun_kwargs)

        if "bulk" in kwargs.keys():
            W += volumetric(J, kwargs["bulk"])

        return W

    return apply_iso


def volumetric(J, bulk):
    return bulk * (J - 1) ** 2 / 2


def displacement_pressure_split(fun):
    """Apply the (u/p)-framework on top of a material formulation (a function
    of the deformation gradient which returns the first Piola-Kirchhoff stress)
    . The additional hydrostatic stress variable `p` is attached as an
    attribute `fun.p` to the augmented function."""

    p = Variable("p")

    @wraps(fun)
    def apply_up(*args, **kwargs):

        F = args[0][0]

        f = fun(*args, **kwargs)

        # check if function is list or tuple
        if not (isinstance(f, list) or isinstance(f, tuple)):
            f = [f]

        P = f[0]
        P_vol = trace(P @ F.T) / det(F) / 3
        K = trace(gradient(P_vol, F) @ F.T) / det(F) / 3

        return [P - (P_vol - p) * cof(F), (P_vol - p) / K, *f[1:]]

    apply_up.p = p

    return apply_up



from ._helpers import isochoric_volumetric_split
from ..math import (
    transpose,
    det,
    DM,
    exp,
    log,
    cos,
    sin,
    pi,
    sqrt,
    if_else,
    invariants,
    trace,
)


@isochoric_volumetric_split
def fiber(F, E, angle, k=1, axis=2, compression=False):
    "Fiber"

    a = angle * pi / 180
    plane = [(1, 2), (2, 0), (0, 1)][axis]

    N = DM.zeros(3)
    N[plane, :] = DM([cos(a), sin(a)])

    C = transpose(F) @ F

    stretch = sqrt((transpose(N) @ C @ N))[0, 0]

    if k == 0:
        strain = log(stretch)
    else:
        strain = 1 / k * (stretch**k - 1)

    if not compression:
        strain = if_else(strain < 0, 0, strain)

    return E * strain**2 / 2


@isochoric_volumetric_split
def fiber_family(F, E, angle, k=1, axis=2, compression=False):
    "Fiber Family"

    f1 = fiber(F, E, angle=angle, k=k, axis=axis, compression=compression)
    f2 = fiber(F, E, angle=-angle, k=k, axis=axis, compression=compression)

    return f1 + f2


@isochoric_volumetric_split
def holzapfel_gasser_ogden(F, c, k1, k2, kappa, angle, axis=2):
    "Holzapfel-Gasser-Ogden"

    C = transpose(F) @ F
    I1, I2, I3 = invariants(C)

    alpha = angle * pi / 180
    plane = [(1, 2), (2, 0), (0, 1)][axis]

    N1 = DM.zeros(3)
    N1[plane, :] = DM([cos(alpha), sin(alpha)])

    N2 = DM.zeros(3)
    N2[plane, :] = DM([cos(alpha), -sin(alpha)])

    A1 = N1 @ transpose(N1)
    A2 = N2 @ transpose(N2)

    I4 = trace(C @ A1)
    I6 = trace(C @ A2)

    W_iso = c / 2 * (I1 - 3)

    w4 = exp(k2 * (kappa * I1 + (1 - 3 * kappa) * I4 - 1) ** 2) - 1
    w6 = exp(k2 * (kappa * I1 + (1 - 3 * kappa) * I6 - 1) ** 2) - 1

    W_aniso = k1 / (2 * k2) * (w4 + w6)

    return W_iso + W_aniso



import numpy as np

from matadi import Variable, Material, MaterialTensor
from matadi.math import det, transpose, trace, invariants, sqrt, dev, ddot


def neohooke(x, mu=1.0, bulk=200.0):
    """Strain energy density function of nearly-incompressible
    Neo-Hookean isotropic hyperelastic material formulation."""

    F = x[0]
    C = transpose(F) @ F

    I1, I2, I3 = invariants(C)
    J = sqrt(I3)

    I1_iso = I3 ** (-1 / 3) * trace(C)

    return mu * (I1_iso - 3) + bulk * (J - 1) ** 2 / 2


def nh_aslist(x, mu=1.0, bulk=200.0):
    return [neohooke(x, mu, bulk)]


def test_simple():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.random.rand(3, 3, 8, 1000)
    for a in range(3):
        FF[a, a] += 1

    # init Material
    W = Material(x=[F], fun=nh_aslist, kwargs={"mu": 1.0, "bulk": 10.0})
    W = Material(x=[F], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0})

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    # dW and DW are always lists...
    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)

    # check output of parallel evaluation
    W0 = W.function([FF], threads=2)
    dW = W.gradient([FF], threads=2)
    DW = W.hessian([FF], threads=2)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (3, 3, 8, 1000)
    assert DW[0].shape == (3, 3, 3, 3, 8, 1000)

    # test single value input
    W00 = W.function([FF[:, :, 0, 0]])

    assert W00[0].shape == (1, 1)

    # init Material with compress
    W = Material(x=[F], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0}, compress=True)
    W0 = W.function([FF])

    assert W0[0].shape == (8, 1000)

    # test single value input
    W00 = W.function([FF[:, :, 0, 0]])

    assert W00[0].shape == (1,)


def test_tensor():

    # variables
    F = Variable("F", 3, 3)
    p = Variable("p", 1, 1)
    z = Variable("z", 5, 16)

    # data
    FF = np.random.rand(3, 3, 8, 1000)
    pp = np.random.rand(8, 1000)
    zz = np.random.rand(5, 16, 8, 1000)

    for a in range(3):
        FF[a, a] += 1

    # init Material
    for fun in [lambda x: dev(x[0]) + ddot(x[0], x[0]), lambda x: x]:

        W = MaterialTensor(x=[F], fun=fun)

        W0 = W.gradient([FF])
        dW = W.hessian([FF])
        # dW and DW are always lists...
        assert W0[0].shape == (3, 3, 8, 1000)
        assert dW[0].shape == (3, 3, 3, 3, 8, 1000)

        # check output of parallel evaluation
        W0 = W.gradient([FF], threads=2)
        dW = W.hessian([FF], threads=2)

        assert W0[0].shape == (3, 3, 8, 1000)
        assert dW[0].shape == (3, 3, 3, 3, 8, 1000)

    # init Material
    pp = np.random.rand(8, 1000)

    W = MaterialTensor(x=[p], fun=lambda x: x[0], compress=True)
    W0 = W.gradient([pp], threads=2)

    assert W0[0].shape == (8, 1000)

    pp = np.random.rand(1, 1, 8, 1000)

    W = MaterialTensor(x=[p], fun=lambda x: x[0])
    W0 = W.gradient([pp])

    assert W0[0].shape == (1, 1, 8, 1000)

    # init mixed Material with upper triangle gradient
    W = MaterialTensor(x=[F, p], fun=lambda x: x, triu=True)
    P = W.gradient([FF, pp])
    A = W.hessian([FF, pp])

    assert len(P) == 2
    assert len(A) == 3

    # init mixed Material with full gradient
    W = MaterialTensor(x=[F, p], fun=lambda x: x, triu=False)
    P = W.gradient([FF, pp])
    A = W.hessian([FF, pp])

    assert len(P) == 2
    assert len(A) == 4

    # init mixed Material with upper triangle gradient and state variables
    W = MaterialTensor(x=[F, p, z], fun=lambda x: [*x, z], triu=True, statevars=1)
    P = W.gradient([FF, pp, zz])
    A = W.hessian([FF, pp, zz])

    assert len(P) == 4
    assert len(A) == 3


if __name__ == "__main__":
    test_simple()
    test_tensor()



from ..._helpers import isochoric_volumetric_split, displacement_pressure_split
from ....math import transpose, sum1, diag, sqrt, inv, det, reshape, dev
from ..quadrature._bazant_oh import BazantOh


@isochoric_volumetric_split
def microsphere_affine_stretch(F, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    C = transpose(F) @ F
    affine_stretch = sqrt(diag(r.T @ C @ r))

    return sum1(f(affine_stretch, **kwargs) * w)


@isochoric_volumetric_split
def microsphere_affine_tube(F, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine area-stretch part."

    r = quadrature.points
    w = quadrature.weights

    Fs = det(F) * transpose(inv(F))
    Cs = transpose(Fs) @ Fs
    affine_areastretch = sqrt(diag(r.T @ Cs @ r))

    return sum1(f(affine_areastretch, **kwargs) * w)


@displacement_pressure_split
def microsphere_affine_force(x, f, *args, **kwargs):
    """Micro-sphere model: Forces of affine stretch model as first Piola-
    Kirchhoff stress tensor embedded into a (u/p)-framework."""

    # sphere quadrature
    sphere = BazantOh(n=21)

    # extract current and initial deformation gradient and state variables
    F = x[0]
    statevars_n = x[-1]

    # volume ratios
    J = det(F)

    # unimodular part of current and initial right Cauchy-Green deformation tensor
    C = F.T @ F
    CG = J ** (-2 / 3) * (C)

    # affine stretches
    lam = sqrt(diag(sphere.points.T @ CG @ sphere.points))

    bulk = kwargs.pop("bulk")

    # fiber forces and state variable update
    f, statevars = f(lam, statevars_n, *args, **kwargs)

    # Second Piola-Kirchhoff stress tensor
    SG = reshape(sum1(f / lam * sphere.weights * sphere.bases), 3, 3)
    S = dev(SG @ CG) @ inv(C) + bulk * (J - 1) * J * inv(C)

    return F @ S, statevars



import numpy as np

from matadi import MaterialTensor, Variable
from matadi.models.microsphere import affine


def nh(stretch, statevars_n, mu=1.0):
    "1d-Linear force-stretch consitututive material formulation."

    return 3 * mu * stretch, statevars_n


def test_microsphere_force():

    F = Variable("F", 3, 3)
    Fn = Variable("Fn", 3, 3)
    Zn = Variable("Zn", 5, 21)

    umat = MaterialTensor(
        x=[F, affine.force.p, Zn],
        fun=affine.force,
        kwargs={"f": nh, "mu": 1.0, "bulk": 5000},
        statevars=1,
        triu=True,
    )

    F = np.random.rand(3, 3, 8, 100) / 2
    p = np.random.rand(1, 8, 100)
    Zn = np.random.rand(5, 21, 8, 100)

    for a in range(3):
        F[a, a] += 1
        Fn[a, a] += 1

    P, Q, Z = umat.function([F, p, Zn])
    A = umat.gradient([F, p, Zn])

    assert P.shape == (3, 3, 8, 100)
    assert Q.shape == (1, 1, 8, 100)
    assert Z.shape == (5, 21, 8, 100)

    assert len(A) == 3


if __name__ == "__main__":
    test_microsphere_force()



from ._misc import morph
from ._hyperelasticity_isotropic import neo_hooke
from ._pseudo_elasticity import ogden_roxburgh
from ..math import gradient
from .._templates import MaterialTensorGeneral


class NeoHookeOgdenRoxburgh(MaterialTensorGeneral):
    "Neo-Hooke and Ogden-Roxburgh material formulation."

    def __init__(self, C10=0.5, r=3, m=1, beta=0):
        def fun(x, C10, r, m, beta):

            # split `x` into the deformation gradient and the state variable
            F, Wmaxn = x[0], x[-1]

            # isochoric and volumetric parts of the hyperelastic
            # strain energy function
            W = neo_hooke(F, C10)

            # pseudo-elastic softening function
            eta, Wmax = ogden_roxburgh(W, Wmaxn, r, m, beta)

            # first Piola-Kirchhoff stress and updated state variable
            # for a pseudo-elastic material formulation
            return eta * gradient(W, F), Wmax

        super().__init__(
            fun=fun, statevars_shape=(1, 1), C10=C10, r=r, m=m, beta=beta
        )


class Morph(MaterialTensorGeneral):
    "MORPH consitutive material formulation."

    def __init__(
        self,
        p1=0.035,
        p2=0.37,
        p3=0.17,
        p4=2.4,
        p5=0.01,
        p6=6.4,
        p7=5.5,
        p8=0.24,
    ):
        def fun(x, p1, p2, p3, p4, p5, p6, p7, p8):

            P, statevars = morph(x, p1, p2, p3, p4, p5, p6, p7, p8)

            return P, statevars

        super().__init__(
            fun=fun,
            statevars_shape=(13, 1),
            p1=p1,
            p2=p2,
            p3=p3,
            p4=p4,
            p5=p5,
            p6=p6,
            p7=p7,
            p8=p8,
        )



import numpy as np

from matadi import Variable, Material, MaterialTensor
from matadi.math import transpose, eigvals, sum1, trace, cof, inv, det, SX, mexp


def fun(x):
    F = x[0]
    return (sum1(eigvals(transpose(F) @ F))[0, 0] - 3) / 2


def fun_mexp(x):
    F = x[0]
    return trace(mexp(transpose(F) @ F))[0, 0]


def test_eigvals():

    # test several repeated principal stretches
    for w in [1, 1.1, 0.1, 2.4, 12]:

        # variables
        F = Variable("F", 3, 3)

        # data
        np.random.seed(2345537)
        FF = np.random.rand(3, 3, 5, 100)
        for a in range(3):
            FF[a, a] += 1

        # input with repeated equal eigenvalues
        FF[:, :, 0, 1] = np.diag(w * np.ones(3))
        FF[:, :, 0, 2] = np.diag([w, w * 1.01, w])

        # init Material
        W = Material(x=[F], fun=fun)

        WW = W.function([FF])
        dW = W.gradient([FF])
        DW = W.hessian([FF])

        Eye = np.eye(3)
        Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

        # check function
        f = FF[:, :, 0, 0]
        c = f.T @ f
        assert np.isclose(WW[0][0, 0, 0, 0], (np.linalg.eigvals(c).sum() - 3) / 2)

        # check gradient
        assert np.allclose(dW[0][:, :, 0, 0], FF[:, :, 0, 0])
        assert np.allclose(dW[0][:, :, 0, 1], FF[:, :, 0, 1])
        assert np.allclose(dW[0][:, :, 0, 2], FF[:, :, 0, 2])

        # check hessian
        assert np.allclose(DW[0][:, :, :, :, 0, 0], Eye4)
        assert np.allclose(DW[0][:, :, :, :, 0, 1], Eye4)
        assert np.allclose(DW[0][:, :, :, :, 0, 2], Eye4)


def test_eigvals_single():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.diag(2.4 * np.ones(3))

    # init Material
    W = Material(x=[F], fun=fun)

    WW = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    Eye = np.eye(3)
    Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

    # check function
    f = FF
    c = f.T @ f
    assert np.isclose(WW[0], (np.linalg.eigvals(c).sum() - 3) / 2)

    # check gradient
    assert np.allclose(dW[0], FF)

    # check hessian
    assert np.allclose(DW[0], Eye4)


def test_cof():

    # variables
    F = Variable("F", 3, 3)

    # data
    FF = np.diag(2.4 * np.ones(3)).reshape(3, 3, 1, 1)
    dF = np.diag(0.2 * np.ones(3)).reshape(3, 3, 1, 1)

    # fun
    def g(x):
        F = x[0]
        return cof(F)

    # init Material
    W = MaterialTensor(x=[F], fun=g)
    W = MaterialTensor(x=[F], fun=lambda x: x[0])

    Eye = np.eye(3)
    Eye4 = np.einsum("ij,kl->ikjl", Eye, Eye)

    WW = W.gradient([FF])
    dW = W.hessian([FF])

    dW_gvp = W.gradient_vector_product([FF, dF])

    assert np.allclose(np.einsum("ij...,ij...->...", dW[0], dF), dW_gvp)

    assert np.allclose(dW[0][:, :, :, :, 0, 0], Eye4)


def test_mexp():

    # test several repeated principal stretches
    for w in [1, 1.1, 0.1, 2.4, 12]:

        # variables
        F = Variable("F", 3, 3)

        # data
        FF = np.diag([w * 1.01, w, w])
        FF = FF.reshape(3, 3, 1, 1)

        # init Material
        W = Material(x=[F], fun=fun_mexp)

        WW = W.function([FF])
        dW = W.gradient([FF])
        DW = W.hessian([FF])

        assert not np.any(np.isnan(WW))
        assert not np.any(np.isnan(dW))
        assert not np.any(np.isnan(DW))


if __name__ == "__main__":

    # test several repeated principal stretches
    test_eigvals()
    test_mexp()

    test_eigvals_single()
    test_cof()
